﻿// Copyright 2020 faddenSoft. All Rights Reserved.
// See the LICENSE.txt file for distribution terms (Apache 2.0).

using System;
using System.Collections.Generic;
using System.ComponentModel;
using PluginCommon;

namespace Apple2_ABM {
    /// <summary>
    /// ABM uses inline data for text output and floating-point math.  Life is much easier
    /// when the data is identified and formatted automatically.
    /// </summary>
    public class InlineCallFormatter : MarshalByRefObject, IPlugin, IPlugin_SymbolList,
            IPlugin_InlineJsr {
        private IApplication mAppRef;
        private byte[] mFileData;

        private const string ADDR_PREFIX = "InA_";
        private const string POSN_PREFIX = "InXY_";
        private const string STRING_PREFIX = "InS_";
        private enum InlineKind { Unknown = 0, Address, Position, String };

        private Dictionary<int, InlineKind> mInlineLabels = new Dictionary<int, InlineKind>();

        public string Identifier {
            get {
                return "ABM inline call formatter";
            }
        }

        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;

            mAppRef.DebugLog("InlineNullStringHandler(id=" +
                AppDomain.CurrentDomain.Id + "): prepare()");
        }

        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
        }

        public void UpdateSymbolList(List<PlSymbol> plSyms) {
            mInlineLabels.Clear();

            foreach (PlSymbol sym in plSyms) {
                if (sym.Label.StartsWith(ADDR_PREFIX)) {
                    mInlineLabels.Add(sym.Value, InlineKind.Address);
                } else if (sym.Label.StartsWith(POSN_PREFIX)) {
                    mInlineLabels.Add(sym.Value, InlineKind.Position);
                } else if (sym.Label.StartsWith(STRING_PREFIX)) {
                    mInlineLabels.Add(sym.Value, InlineKind.String);
                }
            }
            mAppRef.DebugLog("Prefixes matched " + mInlineLabels.Count + " labels");
        }
        public bool IsLabelSignificant(string beforeLabel, string afterLabel) {
            return DoesPrefixMatch(beforeLabel) || DoesPrefixMatch(afterLabel);
        }

        private static bool DoesPrefixMatch(string label) {
            return (label.StartsWith(ADDR_PREFIX) ||
                label.StartsWith(STRING_PREFIX) ||
                label.StartsWith(POSN_PREFIX));
        }

        public void CheckJsr(int offset, int operand, out bool noContinue) {
            noContinue = false;

            InlineKind kind;
            if (!mInlineLabels.TryGetValue(operand, out kind)) {
                return;
            }

            switch (kind) {
                case InlineKind.Address:
                    // Pointer to data.  Format as address.
                    if (!Util.IsInBounds(mFileData, offset + 3, 2)) {
                        return;
                    }
                    mAppRef.SetInlineDataFormat(offset + 3, 2,
                        DataType.NumericLE, DataSubType.Address, null);
                    break;
                case InlineKind.Position:
                    // Position on text screen.  Format as two decimal values.
                    if (!Util.IsInBounds(mFileData, offset + 3, 2)) {
                        return;
                    }
                    mAppRef.SetInlineDataFormat(offset + 3, 1,
                        DataType.NumericLE, DataSubType.Decimal, null);
                    mAppRef.SetInlineDataFormat(offset + 4, 1,
                        DataType.NumericLE, DataSubType.Decimal, null);
                    break;
                case InlineKind.String:
                    // Null-terminated ASCII string.
                    FormatNullTermString(offset);
                    break;
            }
        }

        private void FormatNullTermString(int offset) {
            // search for the terminating null byte
            int nullOff = offset + 3;
            while (nullOff < mFileData.Length) {
                if (mFileData[nullOff] == 0) {
                    break;
                }
                nullOff++;
            }
            if (nullOff == mFileData.Length) {
                mAppRef.DebugLog("Unable to find end of null-terminated string at +" +
                    (offset + 3).ToString("x6"));
                return;
            }

            // Assuming ASCII.  This can be hard-coded, use auto-detection, or look
            // up a value in a project constant.
            mAppRef.SetInlineDataFormat(offset + 3, nullOff - (offset + 3) + 1,
                DataType.StringNullTerm, DataSubType.HighAscii, null);
        }
    }
}
